@@ -13,6 +13,7 @@
 #include "queues.h"
 #include "settings.h"
 #include "utils.h"
+#include "rules.h"
 
 #define FDN_PATH "/org/freedesktop/Notifications"
 #define FDN_IFAC "org.freedesktop.Notifications"
@@ -74,17 +75,31 @@ static const char *introspection_xml =
 
     "        <method name=\"ContextMenuCall\"       />"
     "        <method name=\"NotificationAction\">"
-    "            <arg name=\"number\"     type=\"i\"/>"
+    "            <arg name=\"number\"     type=\"u\"/>"
     "        </method>"
     "        <method name=\"NotificationCloseLast\" />"
     "        <method name=\"NotificationCloseAll\"  />"
+    "        <method name=\"NotificationListHistory\">"
+    "            <arg direction=\"out\" name=\"notifications\"   type=\"aa{sv}\"/>"
+    "        </method>"
+    "        <method name=\"NotificationPopHistory\">"
+    "            <arg direction=\"in\"  name=\"id\"              type=\"u\"/>"
+    "        </method>"
     "        <method name=\"NotificationShow\"      />"
+    "        <method name=\"RuleEnable\">"
+    "            <arg name=\"name\"     type=\"s\"/>"
+    "            <arg name=\"state\"    type=\"i\"/>"
+    "        </method>"
     "        <method name=\"Ping\"                  />"
 
     "        <property name=\"paused\" type=\"b\" access=\"readwrite\">"
     "            <annotation name=\"org.freedesktop.DBus.Property.EmitsChangedSignal\" value=\"true\"/>"
     "        </property>"
 
+    "        <property name=\"displayedLength\" type=\"u\" access=\"read\" />"
+    "        <property name=\"historyLength\" type=\"u\" access=\"read\" />"
+    "        <property name=\"waitingLength\" type=\"u\" access=\"read\" />"
+
     "    </interface>"
     "</node>";
 
@@ -157,15 +172,21 @@ DBUS_METHOD(dunst_ContextMenuCall);
 DBUS_METHOD(dunst_NotificationAction);
 DBUS_METHOD(dunst_NotificationCloseAll);
 DBUS_METHOD(dunst_NotificationCloseLast);
+DBUS_METHOD(dunst_NotificationListHistory);
+DBUS_METHOD(dunst_NotificationPopHistory);
 DBUS_METHOD(dunst_NotificationShow);
+DBUS_METHOD(dunst_RuleEnable);
 DBUS_METHOD(dunst_Ping);
 static struct dbus_method methods_dunst[] = {
-        {"ContextMenuCall",        dbus_cb_dunst_ContextMenuCall},
-        {"NotificationAction",     dbus_cb_dunst_NotificationAction},
-        {"NotificationCloseAll",   dbus_cb_dunst_NotificationCloseAll},
-        {"NotificationCloseLast",  dbus_cb_dunst_NotificationCloseLast},
-        {"NotificationShow",       dbus_cb_dunst_NotificationShow},
-        {"Ping",                   dbus_cb_dunst_Ping},
+        {"ContextMenuCall",           dbus_cb_dunst_ContextMenuCall},
+        {"NotificationAction",        dbus_cb_dunst_NotificationAction},
+        {"NotificationCloseAll",      dbus_cb_dunst_NotificationCloseAll},
+        {"NotificationCloseLast",     dbus_cb_dunst_NotificationCloseLast},
+        {"NotificationListHistory",   dbus_cb_dunst_NotificationListHistory},
+        {"NotificationPopHistory",    dbus_cb_dunst_NotificationPopHistory},
+        {"NotificationShow",          dbus_cb_dunst_NotificationShow},
+        {"Ping",                      dbus_cb_dunst_Ping},
+        {"RuleEnable",                dbus_cb_dunst_RuleEnable},
 };
 
 void dbus_cb_dunst_methods(GDBusConnection *connection,
@@ -210,12 +231,12 @@ static void dbus_cb_dunst_NotificationAction(GDBusConnection *connection,
                                              GVariant *parameters,
                                              GDBusMethodInvocation *invocation)
 {
-        int notification_nr = 0;
-        g_variant_get(parameters, "(i)", &notification_nr);
+        guint32 notification_nr = 0;
+        g_variant_get(parameters, "(u)", &notification_nr);
 
         LOG_D("CMD: Calling action for notification %d", notification_nr);
 
-        if (notification_nr < 0 || queues_length_waiting() < notification_nr) {
+        if (queues_length_waiting() < notification_nr) {
                 g_dbus_method_invocation_return_error(invocation,
                         G_DBUS_ERROR,
                         G_DBUS_ERROR_INVALID_ARGS,
@@ -278,6 +299,148 @@ static void dbus_cb_dunst_NotificationShow(GDBusConnection *connection,
         g_dbus_connection_flush(connection, NULL, NULL, NULL);
 }
 
+static void dbus_cb_dunst_NotificationListHistory(GDBusConnection *connection,
+                                           const gchar *sender,
+                                           GVariant *parameters,
+                                           GDBusMethodInvocation *invocation)
+{
+        LOG_D("CMD: Listing all notifications from history");
+
+        GVariant *answer = NULL;
+        GVariantBuilder *builder;
+
+        builder = g_variant_builder_new(G_VARIANT_TYPE("aa{sv}"));
+
+        GList *notification_list = queues_get_history();
+
+        // reverse chronological list
+        for(int i = queues_length_history(); i > 0; i--) {
+                struct notification *n;
+                n = g_list_nth_data(notification_list, i-1);
+
+                GVariantBuilder n_builder;
+
+                g_variant_builder_init(&n_builder, g_variant_type_new("a{sv}"));
+
+                char *body, *msg, *summary, *appname, *category;
+                char *default_action_name, *icon_path;
+
+                body      = (n->body      == NULL) ? "" : n->body;
+                msg       = (n->msg       == NULL) ? "" : n->msg;
+                summary   = (n->summary   == NULL) ? "" : n->summary;
+                appname   = (n->appname   == NULL) ? "" : n->appname;
+                category  = (n->category  == NULL) ? "" : n->category;
+                default_action_name= (n->default_action_name == NULL) ?
+                        "" : n->default_action_name;
+                icon_path = (n->icon_path == NULL) ? "" : n->icon_path;
+
+                g_variant_builder_add(&n_builder, "{sv}", "body",
+                        g_variant_new_from_bytes(G_VARIANT_TYPE("s"),
+                        g_bytes_new(body, strlen(body)+1), TRUE));
+                g_variant_builder_add(&n_builder, "{sv}", "message",
+                        g_variant_new_from_bytes(G_VARIANT_TYPE("s"),
+                        g_bytes_new(msg, strlen(msg)+1), TRUE));
+                g_variant_builder_add(&n_builder, "{sv}", "summary",
+                        g_variant_new_from_bytes(G_VARIANT_TYPE("s"),
+                        g_bytes_new(summary, strlen(summary)+1), TRUE));
+                g_variant_builder_add(&n_builder, "{sv}", "appname",
+                        g_variant_new_from_bytes(G_VARIANT_TYPE("s"),
+                        g_bytes_new(appname, strlen(appname)+1), TRUE));
+                g_variant_builder_add(&n_builder, "{sv}", "category",
+                        g_variant_new_from_bytes(G_VARIANT_TYPE("s"),
+                        g_bytes_new(category, strlen(category)+1), TRUE));
+                g_variant_builder_add(&n_builder, "{sv}", "default_action_name",
+                        g_variant_new_from_bytes(G_VARIANT_TYPE("s"),
+                        g_bytes_new(default_action_name,
+                        strlen(default_action_name)+1), TRUE));
+                g_variant_builder_add(&n_builder, "{sv}", "icon_path",
+                        g_variant_new_from_bytes(G_VARIANT_TYPE("s"),
+                        g_bytes_new(icon_path, strlen(icon_path)+1), TRUE));
+                g_variant_builder_add(&n_builder, "{sv}", "id",
+                        g_variant_new_from_bytes(G_VARIANT_TYPE("i"),
+                        g_bytes_new(&n->id, sizeof(int)), TRUE));
+                g_variant_builder_add(&n_builder, "{sv}", "timestamp",
+                        g_variant_new_from_bytes(G_VARIANT_TYPE("x"),
+                        g_bytes_new(&n->timestamp, sizeof(gint64)), TRUE));
+                g_variant_builder_add(&n_builder, "{sv}", "timeout",
+                        g_variant_new_from_bytes(G_VARIANT_TYPE("x"),
+                        g_bytes_new(&n->timeout, sizeof(gint64)), TRUE));
+                g_variant_builder_add(&n_builder, "{sv}", "progress",
+                        g_variant_new_from_bytes(G_VARIANT_TYPE("i"),
+                        g_bytes_new(&n->progress, sizeof(int)), TRUE));
+
+                g_variant_builder_add(builder, "a{sv}", &n_builder);
+
+        }
+
+        answer = g_variant_new("(aa{sv})", builder);
+
+        g_clear_pointer(&builder, g_variant_builder_unref);
+        g_dbus_method_invocation_return_value(invocation, answer);
+        g_dbus_connection_flush(connection, NULL, NULL, NULL);
+}
+
+static void dbus_cb_dunst_NotificationPopHistory(GDBusConnection *connection,
+                                           const gchar *sender,
+                                           GVariant *parameters,
+                                           GDBusMethodInvocation *invocation)
+{
+        LOG_D("CMD: Popping notification from history");
+
+        guint32 id;
+        g_variant_get(parameters, "(u)", &id);
+
+        queues_history_pop_by_id(id);
+        wake_up();
+
+        g_dbus_method_invocation_return_value(invocation, NULL);
+        g_dbus_connection_flush(connection, NULL, NULL, NULL);
+}
+
+static void dbus_cb_dunst_RuleEnable(GDBusConnection *connection,
+                                     const gchar *sender,
+                                     GVariant *parameters,
+                                     GDBusMethodInvocation *invocation)
+{
+        // dbus param state: 0 → disable, 1 → enable, 2 → toggle.
+
+        int state = 0;
+        char *name = NULL;
+        g_variant_get(parameters, "(si)", &name, &state);
+
+        LOG_D("CMD: Changing rule \"%s\" enable state to %d", name, state);
+
+        if (state < 0 || state > 2) {
+                g_dbus_method_invocation_return_error(invocation,
+                        G_DBUS_ERROR,
+                        G_DBUS_ERROR_INVALID_ARGS,
+                        "Couldn't understand state %d. It must be 0, 1 or 2",
+                        state);
+                return;
+        }
+
+        struct rule *target_rule = get_rule(name);
+
+        if (target_rule == NULL) {
+                g_dbus_method_invocation_return_error(invocation,
+                        G_DBUS_ERROR,
+                        G_DBUS_ERROR_INVALID_ARGS,
+                        "There is no rule named \"%s\"",
+                        name);
+                return;
+        }
+
+        if (state == 0)
+                target_rule->enabled = false;
+        else if (state == 1)
+                target_rule->enabled = true;
+        else if (state == 2)
+                target_rule->enabled = !target_rule->enabled;
+
+        g_dbus_method_invocation_return_value(invocation, NULL);
+        g_dbus_connection_flush(connection, NULL, NULL, NULL);
+}
+
 /* Just a simple Ping command to give the ability to dunstctl to test for the existence of this interface
  * Any other way requires parsing the XML of the Introspection or other foo. Just calling the Ping on an old dunst version will fail. */
 static void dbus_cb_dunst_Ping(GDBusConnection *connection,
@@ -303,11 +466,15 @@ static void dbus_cb_GetCapabilities(
         g_variant_builder_add(builder, "s", "actions");
         g_variant_builder_add(builder, "s", "body");
         g_variant_builder_add(builder, "s", "body-hyperlinks");
+        g_variant_builder_add(builder, "s", "icon-static");
 
         for (int i = 0; i < sizeof(stack_tag_hints)/sizeof(*stack_tag_hints); ++i)
                 g_variant_builder_add(builder, "s", stack_tag_hints[i]);
 
-        if (settings.markup != MARKUP_NO)
+        // Since markup isn't a global variable anymore, look it up in the
+        // global rule
+        struct rule *global_rule = get_rule("global");
+        if (global_rule && global_rule->markup != MARKUP_NO)
                 g_variant_builder_add(builder, "s", "body-markup");
 
         value = g_variant_new("(as)", builder);
@@ -361,26 +528,13 @@ static struct notification *dbus_message_to_notification(const gchar *sender, GV
         }
 
         GVariant *dict_value;
+
+        // First process the items that can be filtered on
         if ((dict_value = g_variant_lookup_value(hints, "urgency", G_VARIANT_TYPE_BYTE))) {
                 n->urgency = g_variant_get_byte(dict_value);
                 g_variant_unref(dict_value);
         }
 
-        if ((dict_value = g_variant_lookup_value(hints, "fgcolor", G_VARIANT_TYPE_STRING))) {
-                n->colors.fg = g_variant_dup_string(dict_value, NULL);
-                g_variant_unref(dict_value);
-        }
-
-        if ((dict_value = g_variant_lookup_value(hints, "bgcolor", G_VARIANT_TYPE_STRING))) {
-                n->colors.bg = g_variant_dup_string(dict_value, NULL);
-                g_variant_unref(dict_value);
-        }
-
-        if ((dict_value = g_variant_lookup_value(hints, "frcolor", G_VARIANT_TYPE_STRING))) {
-                n->colors.frame = g_variant_dup_string(dict_value, NULL);
-                g_variant_unref(dict_value);
-        }
-
         if ((dict_value = g_variant_lookup_value(hints, "category", G_VARIANT_TYPE_STRING))) {
                 n->category = g_variant_dup_string(dict_value, NULL);
                 g_variant_unref(dict_value);
@@ -391,20 +545,26 @@ static struct notification *dbus_message_to_notification(const gchar *sender, GV
                 g_variant_unref(dict_value);
         }
 
-        if ((dict_value = g_variant_lookup_value(hints, "image-path", G_VARIANT_TYPE_STRING))) {
-                g_free(n->iconname);
-                n->iconname = g_variant_dup_string(dict_value, NULL);
+        if ((dict_value = g_variant_lookup_value(hints, "value", G_VARIANT_TYPE_INT32))) {
+                n->progress = g_variant_get_int32(dict_value);
+                g_variant_unref(dict_value);
+        } else if ((dict_value = g_variant_lookup_value(hints, "value", G_VARIANT_TYPE_UINT32))) {
+                n->progress = g_variant_get_uint32(dict_value);
                 g_variant_unref(dict_value);
         }
+        if (n->progress < 0)
+                n->progress = -1;
 
-        dict_value = g_variant_lookup_value(hints, "image-data", G_VARIANT_TYPE("(iiibiiay)"));
-        if (!dict_value)
-                dict_value = g_variant_lookup_value(hints, "image_data", G_VARIANT_TYPE("(iiibiiay)"));
-        if (!dict_value)
-                dict_value = g_variant_lookup_value(hints, "icon_data", G_VARIANT_TYPE("(iiibiiay)"));
-        if (dict_value) {
-                notification_icon_replace_data(n, dict_value);
-                g_variant_unref(dict_value);
+        /* Check for hints that define the stack_tag
+         *
+         * Only accept to first one we find.
+         */
+        for (int i = 0; i < sizeof(stack_tag_hints)/sizeof(*stack_tag_hints); ++i) {
+                if ((dict_value = g_variant_lookup_value(hints, stack_tag_hints[i], G_VARIANT_TYPE_STRING))) {
+                        n->stack_tag = g_variant_dup_string(dict_value, NULL);
+                        g_variant_unref(dict_value);
+                        break;
+                }
         }
 
         /* Check for transient hints
@@ -424,24 +584,53 @@ static struct notification *dbus_message_to_notification(const gchar *sender, GV
                 g_variant_unref(dict_value);
         }
 
-        if ((dict_value = g_variant_lookup_value(hints, "value", G_VARIANT_TYPE_INT32))) {
-                n->progress = g_variant_get_int32(dict_value);
-                g_variant_unref(dict_value);
-        } else if ((dict_value = g_variant_lookup_value(hints, "value", G_VARIANT_TYPE_UINT32))) {
-                n->progress = g_variant_get_uint32(dict_value);
+        if ((dict_value = g_variant_lookup_value(hints, "image-path", G_VARIANT_TYPE_STRING))) {
+                g_free(n->iconname);
+                n->iconname = g_variant_dup_string(dict_value, NULL);
                 g_variant_unref(dict_value);
         }
 
-        /* Check for hints that define the stack_tag
-         *
-         * Only accept to first one we find.
-         */
-        for (int i = 0; i < sizeof(stack_tag_hints)/sizeof(*stack_tag_hints); ++i) {
-                if ((dict_value = g_variant_lookup_value(hints, stack_tag_hints[i], G_VARIANT_TYPE_STRING))) {
-                        n->stack_tag = g_variant_dup_string(dict_value, NULL);
-                        g_variant_unref(dict_value);
-                        break;
-                }
+        // Set raw icon data only after initializing the notification, so the
+        // desired icon size is known. This way the buffer can be immediately
+        // rescaled. If at some point you might want to match by if a
+        // notificaton has an image, this has to be reworked.
+        dict_value = g_variant_lookup_value(hints, "image-data", G_VARIANT_TYPE("(iiibiiay)"));
+        if (!dict_value)
+                dict_value = g_variant_lookup_value(hints, "image_data", G_VARIANT_TYPE("(iiibiiay)"));
+        if (!dict_value)
+                dict_value = g_variant_lookup_value(hints, "icon_data", G_VARIANT_TYPE("(iiibiiay)"));
+        if (dict_value) {
+                notification_icon_replace_data(n, dict_value);
+                g_variant_unref(dict_value);
+        }
+
+        // All attributes that have to be set before initializations are set,
+        // so we can initialize the notification. This applies all rules that
+        // are defined and applies the formatting to the message.
+        notification_init(n);
+
+        // Modify these values after the notification is initialized and all rules are applied.
+        if ((dict_value = g_variant_lookup_value(hints, "fgcolor", G_VARIANT_TYPE_STRING))) {
+                g_free(n->colors.fg);
+                n->colors.fg = g_variant_dup_string(dict_value, NULL);
+                g_variant_unref(dict_value);
+        }
+
+        if ((dict_value = g_variant_lookup_value(hints, "bgcolor", G_VARIANT_TYPE_STRING))) {
+                g_free(n->colors.bg);
+                n->colors.bg = g_variant_dup_string(dict_value, NULL);
+                g_variant_unref(dict_value);
+        }
+
+        if ((dict_value = g_variant_lookup_value(hints, "frcolor", G_VARIANT_TYPE_STRING))) {
+                g_free(n->colors.frame);
+                n->colors.frame = g_variant_dup_string(dict_value, NULL);
+                g_variant_unref(dict_value);
+        }
+
+        if ((dict_value = g_variant_lookup_value(hints, "hlcolor", G_VARIANT_TYPE_STRING))) {
+                n->colors.highlight = g_variant_dup_string(dict_value, NULL);
+                g_variant_unref(dict_value);
         }
 
         if (timeout >= 0)
@@ -451,7 +640,6 @@ static struct notification *dbus_message_to_notification(const gchar *sender, GV
         g_variant_type_free(required_type);
         g_free(actions); // the strv is only a shallow copy
 
-        notification_init(n);
         return n;
 }
 
@@ -496,10 +684,10 @@ static void dbus_cb_CloseNotification(
         g_variant_get(parameters, "(u)", &id);
         if (settings.ignore_dbusclose) {
                 LOG_D("Ignoring CloseNotification message");
-                // Stay commpliant by lying to the sender,  telling him we closed the notification 
+                // Stay commpliant by lying to the sender,  telling him we closed the notification
                 if (id > 0) {
                         struct notification *n = queues_get_by_id(id);
-                        if (n) 
+                        if (n)
                                 signal_notification_closed(n, REASON_SIG);
                 }
         } else {
@@ -556,6 +744,27 @@ void signal_notification_closed(struct notification *n, enum reason reason)
         if (err) {
                 LOG_W("Unable to close notification: %s", err->message);
                 g_error_free(err);
+        } else {
+                char* reason_string;
+                switch (reason) {
+                        case REASON_TIME:
+                                reason_string="time";
+                                break;
+                        case REASON_USER:
+                                reason_string="user";
+                                break;
+                        case REASON_SIG:
+                                reason_string="signal";
+                                break;
+                        case REASON_UNDEF:
+                                reason_string="undfined";
+                                break;
+                        default:
+                                reason_string="unknown";
+                }
+
+                LOG_D("Queues: Closing notification for reason: %s", reason_string);
+
         }
 
 }
@@ -597,6 +806,15 @@ GVariant *dbus_cb_dunst_Properties_Get(GDBusConnection *connection,
 
         if (STR_EQ(property_name, "paused")) {
                 return g_variant_new_boolean(!status.running);
+        } else if (STR_EQ(property_name, "displayedLength")) {
+                unsigned int displayed = queues_length_displayed();
+                return g_variant_new_uint32(displayed);
+        } else if (STR_EQ(property_name, "historyLength")) {
+                unsigned int history =  queues_length_history();
+                return g_variant_new_uint32(history);
+        } else if (STR_EQ(property_name, "waitingLength")) {
+                unsigned int waiting =  queues_length_waiting();
+                return g_variant_new_uint32(waiting);
         } else {
                 LOG_W("Unknown property!\n");
                 *error = g_error_new(G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_PROPERTY, "Unknown property");
@@ -616,6 +834,22 @@ gboolean dbus_cb_dunst_Properties_Set(GDBusConnection *connection,
         if (STR_EQ(property_name, "paused")) {
                 dunst_status(S_RUNNING, !g_variant_get_boolean(value));
                 wake_up();
+
+                GVariantBuilder *builder = g_variant_builder_new(G_VARIANT_TYPE_ARRAY);
+                GVariantBuilder *invalidated_builder = g_variant_builder_new(G_VARIANT_TYPE("as"));
+                g_variant_builder_add(builder,
+                                      "{sv}",
+                                      "paused", g_variant_new_boolean(g_variant_get_boolean(value)));
+                g_dbus_connection_emit_signal(connection,
+                                              NULL,
+                                              object_path,
+                                              "org.freedesktop.DBus.Properties",
+                                              "PropertiesChanged",
+                                              g_variant_new("(sa{sv}as)",
+                                                            interface_name,
+                                                            builder,
+                                                            invalidated_builder),
+                                              NULL);
                 return true;
         }
 
@@ -831,4 +1065,4 @@ void dbus_teardown(int owner_id)
         g_bus_unown_name(owner_id);
 }
 
-/* vim: set tabstop=8 shiftwidth=8 expandtab textwidth=0: */
+/* vim: set ft=c tabstop=8 shiftwidth=8 expandtab textwidth=0: */
