@@ -5,16 +5,13 @@
 #include <gdk-pixbuf/gdk-pixbuf.h>
 #include <stdbool.h>
 #include <string.h>
+#include <math.h>
 
 #include "log.h"
 #include "notification.h"
 #include "settings.h"
 #include "utils.h"
-
-static bool is_readable_file(const char *filename)
-{
-        return (access(filename, R_OK) != -1);
-}
+#include "icon-lookup.h"
 
 /**
  * Reassemble the data parts of a GdkPixbuf into a cairo_surface_t's data field.
@@ -86,9 +83,19 @@ static void pixbuf_data_to_cairo_data(
         }
 }
 
+int get_icon_width(cairo_surface_t *icon, double scale) {
+        return round(cairo_image_surface_get_width(icon) / scale);
+}
+
+int get_icon_height(cairo_surface_t *icon, double scale) {
+        return round(cairo_image_surface_get_height(icon) / scale);
+}
+
 cairo_surface_t *gdk_pixbuf_to_cairo_surface(GdkPixbuf *pixbuf)
 {
-        assert(pixbuf);
+        if (!pixbuf) {
+                return NULL;
+        }
 
         int width  = gdk_pixbuf_get_width(pixbuf);
         int height = gdk_pixbuf_get_height(pixbuf);
@@ -139,23 +146,55 @@ static bool icon_size_clamp(int *w, int *h) {
         return FALSE;
 }
 
+static bool icon_size_clamp2(int *w, int *h, int desired_size, double scale) {
+        int largest_size = MAX(*w, *h);
+        int new_size = desired_size * scale;
+        if (largest_size != new_size) {
+                double scale = (double) new_size / (double) largest_size;
+                *w = round(*w * scale);
+                *h = round(*h * scale);
+                return true;
+        }
+        return false;
+}
 /**
- * Scales the given GdkPixbuf if necessary according to the settings.
+ * Scales the given GdkPixbuf to a given size.. If the image is not square, the
+ * largest size will be scaled up to the given size.
+ *
+ * The icon is scaled to a size of icon_size * dpi_scale.
  *
  * @param pixbuf (nullable) The pixbuf, which may be too big.
  *                          Takes ownership of the reference.
+ * @param icon_size An integer the unscaled icon size.
+ * @param dpi_scale A double for the dpi scaling.
  * @return the scaled version of the pixbuf. If scaling wasn't
  *         necessary, it returns the same pixbuf. Transfers full
  *         ownership of the reference.
  */
-static GdkPixbuf *icon_pixbuf_scale(GdkPixbuf *pixbuf)
+static GdkPixbuf *icon_pixbuf_scale_to_size(GdkPixbuf *pixbuf, int icon_size, double dpi_scale)
 {
         ASSERT_OR_RET(pixbuf, NULL);
 
         int w = gdk_pixbuf_get_width(pixbuf);
         int h = gdk_pixbuf_get_height(pixbuf);
+        bool needs_change = false;
 
-        if (icon_size_clamp(&w, &h)) {
+        if (settings.enable_recursive_icon_lookup)
+        {
+                needs_change = icon_size_clamp2(&w, &h, icon_size, dpi_scale);
+                if (needs_change) {
+                        LOG_D("Scaling to a size of %ix%i", w, h);
+                        LOG_D("While the icon size and scale are %ix%f", icon_size, dpi_scale);
+                }
+        } else {
+                // TODO immediately rescale icon upon scale changes
+                if(icon_size_clamp(&w, &h)) {
+                        w = round(w * dpi_scale);
+                        h = round(h * dpi_scale);
+                        needs_change = true;
+                }
+        }
+        if (needs_change) {
                 GdkPixbuf *scaled = gdk_pixbuf_scale_simple(
                                 pixbuf,
                                 w,
@@ -164,11 +203,10 @@ static GdkPixbuf *icon_pixbuf_scale(GdkPixbuf *pixbuf)
                 g_object_unref(pixbuf);
                 pixbuf = scaled;
         }
-
         return pixbuf;
 }
 
-GdkPixbuf *get_pixbuf_from_file(const char *filename)
+GdkPixbuf *get_pixbuf_from_file(const char *filename, int icon_size, double scale)
 {
         char *path = string_to_path(g_strdup(filename));
         GError *error = NULL;
@@ -179,12 +217,24 @@ GdkPixbuf *get_pixbuf_from_file(const char *filename)
                 g_free(path);
                 return NULL;
         }
-        icon_size_clamp(&w, &h);
-        GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file_at_scale(path,
-                                                              w,
-                                                              h,
-                                                              TRUE,
-                                                              &error);
+        GdkPixbuf *pixbuf = NULL;
+        if (settings.enable_recursive_icon_lookup)
+        {
+                icon_size_clamp2(&w, &h, icon_size, scale);
+                pixbuf = gdk_pixbuf_new_from_file_at_scale(path,
+                                w,
+                                h,
+                                TRUE,
+                                &error);
+        } else {
+                // TODO immediately rescale icon upon scale changes
+                icon_size_clamp(&w, &h);
+                pixbuf = gdk_pixbuf_new_from_file_at_scale(path,
+                                round(w * scale),
+                                round(h * scale),
+                                TRUE,
+                                &error);
+        }
 
         if (error) {
                 LOG_W("%s", error->message);
@@ -195,14 +245,19 @@ GdkPixbuf *get_pixbuf_from_file(const char *filename)
         return pixbuf;
 }
 
-GdkPixbuf *get_pixbuf_from_icon(const char *iconname)
+char *get_path_from_icon_name(const char *iconname, int size)
 {
+        if (settings.enable_recursive_icon_lookup) {
+                char *path = find_icon_path(iconname, size);
+                LOG_I("Found icon at %s", path);
+                return path;
+        }
         if (STR_EMPTY(iconname))
                 return NULL;
 
         const char *suffixes[] = { ".svg", ".svgz", ".png", ".xpm", NULL };
-        GdkPixbuf *pixbuf = NULL;
         gchar *uri_path = NULL;
+        char *new_name = NULL;
 
         if (g_str_has_prefix(iconname, "file://")) {
                 uri_path = g_filename_from_uri(iconname, NULL, NULL);
@@ -212,7 +267,7 @@ GdkPixbuf *get_pixbuf_from_icon(const char *iconname)
 
         /* absolute path? */
         if (iconname[0] == '/' || iconname[0] == '~') {
-                pixbuf = get_pixbuf_from_file(iconname);
+                new_name = g_strdup(iconname);
         } else {
         /* search in icon_path */
                 char *start = settings.icon_path,
@@ -224,41 +279,33 @@ GdkPixbuf *get_pixbuf_from_icon(const char *iconname)
                         current_folder = g_strndup(start, end - start);
 
                         for (const char **suf = suffixes; *suf; suf++) {
-                                maybe_icon_path = g_strconcat(current_folder, "/", iconname, *suf, NULL);
-                                if (is_readable_file(maybe_icon_path))
-                                        pixbuf = get_pixbuf_from_file(maybe_icon_path);
+                                gchar *name_with_extension = g_strconcat(iconname, *suf, NULL);
+                                maybe_icon_path = g_build_filename(current_folder, name_with_extension, NULL);
+                                if (is_readable_file(maybe_icon_path)) {
+                                        new_name = g_strdup(maybe_icon_path);
+                                }
+                                g_free(name_with_extension);
                                 g_free(maybe_icon_path);
 
-                                if (pixbuf)
+                                if (new_name)
                                         break;
                         }
 
                         g_free(current_folder);
-                        if (pixbuf)
+                        if (new_name)
                                 break;
 
                         start = end + 1;
                 } while (STR_FULL(end));
-                if (!pixbuf)
+                if (!new_name)
                         LOG_W("No icon found in path: '%s'", iconname);
         }
 
         g_free(uri_path);
-        return pixbuf;
+        return new_name;
 }
 
-GdkPixbuf *icon_get_for_name(const char *name, char **id)
-{
-        ASSERT_OR_RET(name, NULL);
-        ASSERT_OR_RET(id, NULL);
-
-        GdkPixbuf *pb = get_pixbuf_from_icon(name);
-        if (pb)
-                *id = g_strdup(name);
-        return pb;
-}
-
-GdkPixbuf *icon_get_for_data(GVariant *data, char **id)
+GdkPixbuf *icon_get_for_data(GVariant *data, char **id, double dpi_scale, int icon_size)
 {
         ASSERT_OR_RET(data, NULL);
         ASSERT_OR_RET(id, NULL);
@@ -326,7 +373,13 @@ GdkPixbuf *icon_get_for_data(GVariant *data, char **id)
                 return NULL;
         }
 
+        // g_memdup is deprecated in glib 2.67.4 and higher.
+        // g_memdup2 is a safer alternative
+#if GLIB_CHECK_VERSION(2,67,3)
+        data_pb = (guchar *) g_memdup2(g_variant_get_data(data_variant), len_actual);
+#else
         data_pb = (guchar *) g_memdup(g_variant_get_data(data_variant), len_actual);
+#endif
 
         pixbuf = gdk_pixbuf_new_from_data(data_pb,
                                           GDK_COLORSPACE_RGB,
@@ -362,9 +415,9 @@ GdkPixbuf *icon_get_for_data(GVariant *data, char **id)
         g_free(data_chk);
         g_variant_unref(data_variant);
 
-        pixbuf = icon_pixbuf_scale(pixbuf);
+        pixbuf = icon_pixbuf_scale_to_size(pixbuf, icon_size, dpi_scale);
 
         return pixbuf;
 }
 
-/* vim: set tabstop=8 shiftwidth=8 expandtab textwidth=0: */
+/* vim: set ft=c tabstop=8 shiftwidth=8 expandtab textwidth=0: */
