@@ -2,14 +2,22 @@
 #ifndef DUNST_NOTIFICATION_H
 #define DUNST_NOTIFICATION_H
 
-#include <gdk-pixbuf/gdk-pixbuf.h>
 #include <glib.h>
 #include <stdbool.h>
+#include <pango/pango-layout.h>
+#include <cairo.h>
 
 #include "markup.h"
 
 #define DUNST_NOTIF_MAX_CHARS 50000
 
+enum icon_position {
+        ICON_LEFT,
+        ICON_RIGHT,
+        ICON_TOP,
+        ICON_OFF
+};
+
 enum behavior_fullscreen {
         FS_NULL,      //!< Invalid value
         FS_DELAY,     //!< Delay the notification until leaving fullscreen mode
@@ -33,6 +41,7 @@ struct notification_colors {
         char *frame;
         char *bg;
         char *fg;
+        char *highlight;
 };
 
 struct notification {
@@ -48,17 +57,24 @@ struct notification {
         char *desktop_entry;     /**< The desktop entry hint sent via every GApplication */
         enum urgency urgency;
 
-        GdkPixbuf *icon;         /**< The raw cached icon data used to draw */
-        char *icon_id;           /**< plain icon information, which acts as the pixbuf's id, which is saved in .icon
+        cairo_surface_t *icon;         /**< The raw cached icon data used to draw */
+        char *icon_id;           /**< Plain icon information, which acts as the icon's id.
                                       May be a hash for a raw icon or a name/path for a regular app icon. */
-        char *iconname;          /**< plain icon information (may be a path or just a name)
-                                      Use this to compare the icon name with rules.*/
+        char *iconname;          /**< plain icon information (may be a path or just a name) as recieved from dbus.
+                                   Use this to compare the icon name with rules. May also be modified by rules.*/
+        char *icon_path;         /**< Full path to the notification's icon. */
+        char *default_icon_name; /**< The icon that is used when no other icon is available. */
+        int icon_size;           /**< Size of the icon used for searching the right icon. */
+        enum icon_position icon_position;       /**< Icon position (enum left,right,top,off). */
 
-        gint64 start;      /**< begin of current display */
-        gint64 timestamp;  /**< arrival time */
-        gint64 timeout;    /**< time to display */
+        gint64 start;      /**< begin of current display (in milliseconds) */
+        gint64 timestamp;  /**< arrival time (in milliseconds) */
+        gint64 timeout;    /**< time to display (in milliseconds) */
+        int locked;     /**< If non-zero the notification is locked **/
+        PangoAlignment progress_bar_alignment; /**< Horizontal alignment of the progress bar **/
 
         GHashTable *actions;
+        char *default_action_name; /**< The name of the action to be invoked on do_action */
 
         enum markup_mode markup;
         const char *format;
@@ -81,6 +97,11 @@ struct notification {
         int displayed_height;
         enum behavior_fullscreen fullscreen; //!< The instruction what to do with it, when desktop enters fullscreen
         bool script_run;        /**< Has the script been executed already? */
+        guint8 marked_for_closure;
+        bool word_wrap;
+        PangoEllipsizeMode ellipsize;
+        PangoAlignment alignment;
+        bool hide_text;
 
         /* derived fields */
         char *msg;            /**< formatted message */
@@ -138,6 +159,22 @@ int notification_cmp_data(const void *va, const void *vb, void *data);
 
 bool notification_is_duplicate(const struct notification *a, const struct notification *b);
 
+bool notification_is_locked(struct notification *n);
+
+struct notification *notification_lock(struct notification *n);
+
+struct notification *notification_unlock(struct notification *n);
+
+/**
+ * Transfer the image surface of \p from to \p to. The image surface is
+ * transfered only if the icon names match. When the icon is transferred, it is
+ * removed from the old notification to make sure it's not freed twice.
+ *
+ * @param from The notification of which the icon surface is removed.
+ * @param to The notification that receives the icon surface.
+ */
+void notification_transfer_icon(struct notification *from, struct notification *to);
+
 /**Replace the current notification's icon with the icon specified by path.
  *
  * Removes the reference for the previous icon automatically and will also free the
@@ -186,11 +223,24 @@ void notification_replace_single_field(char **haystack,
 void notification_update_text_to_render(struct notification *n);
 
 /**
- * If the notification has exactly one action, or one is marked as default,
- * invoke it. If there are multiple and no default, open the context menu. If
- * there are no actions, proceed similarly with urls.
+ * If the notification has an action named n->default_action_name or there is only one
+ * action and n->default_action_name is set to "default", invoke it. If there is no
+ * such action, open the context menu if threre are other actions. Otherwise, do nothing.
  */
-void notification_do_action(const struct notification *n);
+void notification_do_action(struct notification *n);
+
+/**
+ * If the notification has exactly one url, invoke it. If there are multiple,
+ * open the context menu. If there are no urls, do nothing.
+ */
+void notification_open_url(struct notification *n);
+
+/**
+ * Open the context menu for the notification.
+ * 
+ * Convenience function that creates the GList and passes it to context_menu_for().
+ */
+void notification_open_context_menu(struct notification *n);
 
 /**
  * Remove all client action data from the notification.
@@ -212,4 +262,4 @@ const char *notification_urgency_to_string(const enum urgency urgency);
 const char *enum_to_string_fullscreen(enum behavior_fullscreen in);
 
 #endif
-/* vim: set tabstop=8 shiftwidth=8 expandtab textwidth=0: */
+/* vim: set ft=c tabstop=8 shiftwidth=8 expandtab textwidth=0: */
